#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <float.h>

#include <strstream.h>

#include "column.h"
#include "head.h"
#include "util.h"

FitsColumn::FitsColumn(FitsHead* head, int i, int off)
{
  index_ = i;
  width_ = 0;
  offset_ = off;
  type_ = ' ';

  tform_ = head->getString(ftcat("TFORM",i));
  ttype_ = head->getString(ftcat("TTYPE",i));
  tunit_ = head->getString(ftcat("TUNIT",i));
  tscal_ = head->getReal(ftcat("TSCAL",i), 1);
  tzero_ = head->getReal(ftcat("TZERO",i), 0);
  hastnull_ = head->find(ftcat("TNULL",i)) ? 1:0;
  tnull_ = head->getInteger(ftcat("TNULL",i), 0);

  char* tl = head->find(ftcat("TLMAX",i));
  char* ta = head->find(ftcat("TALEN",i));
  char* ax = head->find(ftcat("AXLEN",i));

  // this provides backward compatibility

  if (tl) {
    hastlmin_ = head->find(ftcat("TLMIN",i)) ? 1:0;
    hastlmax_ = 1;
    tlmin_ = head->getReal(ftcat("TLMIN",i), 0);
    tlmax_ = head->getReal(ftcat("TLMAX",i), 0);
  }
  else if (ta) {
    hastlmin_ = 0;
    hastlmax_ = 1;
    tlmin_ = 1;
    tlmax_ = head->getReal(ftcat("TALEN",i), 0);
  }
  else if (ax) {
    hastlmin_ = 0;
    hastlmax_ = 1;
    tlmin_ = 1;
    tlmax_ = head->getReal(ftcat("AXLEN",i), 0);
  }
  else {
    hastlmin_ = 0;
    hastlmax_ = 0;
    tlmin_ = 0;
    tlmax_ = 0;
  }
}

FitsColumn::~FitsColumn()
{
  if (tform_)
    delete [] tform_;
  if (tunit_)
    delete [] tunit_;
  if (ttype_)
    delete [] ttype_;
}

FitsAsciiColumn::FitsAsciiColumn(FitsHead* head, int i, int offset)
  : FitsColumn(head, i, offset)
{
  prec_ = 0;

  tbcol_ = head->getInteger(ftcat("TBCOL",i),0);

  // parse TFORM

  if (tform_) {
    istrstream str(tform_);
    str >> type_ >> width_;
    if (type_ == 'F' || type_ == 'E' || type_ == 'D') {
      char s;
      str >> s >> prec_;
    }
  }
}

double FitsAsciiColumn::value(const char* ptr)
{
  istrstream str(ptr+offset_);
  double r;
  str >> r;

  return r;
}

Vector FitsAsciiColumn::dimension()
{
  if (hastlmin_ || hastlmax_)
    return Vector(tlmin_,tlmax_);
  else
    switch (type_) {
    case 'A':
      return Vector(0,UCHAR_MAX);
      break;
    case 'I':
#if defined( __alpha) || defined(__mips64)
      return Vector(INT_MIN,INT_MAX);
#else
      return Vector(LONG_MIN,LONG_MAX);
#endif
      break;
    case 'F':
    case 'E':
      return Vector(FLT_MIN,FLT_MAX);
      break;
    case 'D':
      return Vector(DBL_MIN,DBL_MAX);
      break;
    }
}

FitsBinColumn::FitsBinColumn(FitsHead* head, int i, int offset)
  : FitsColumn(head, i, offset)
{
  repeat_ = 1;
  tdisp_ = head->getString(ftcat("TDISP",i));
}

FitsBinColumn::~FitsBinColumn()
{
  if (tdisp_)
    delete [] tdisp_;
}

FitsBinColumnA::FitsBinColumnA(FitsHead* head, int i, int offset)
  : FitsBinColumn(head, i, offset)
{
  Awidth_ = 0;
  Aterm_ = 0;

  // parse TFORM

  if (tform_) {
    istrstream str(tform_);
    if (isalpha(tform_[0]))
      str >> type_;
    else
      str >> repeat_ >> type_;

    char s;
    str >> s >> s >> s >> s >> s >> Awidth_ >> s >> oct >> Aterm_;
  }

  // and determine field width

  width_ = repeat_ * (Awidth_+1);
}

double FitsBinColumnA::value(const char* ptr)
{
  double r;

  istrstream str(ptr+offset_, width_);
  str >> r;

  return r;
}

char* FitsBinColumnA::str(const char* ptr)
{
  char* r = new char[width_+1];
  strncpy(r,ptr+offset_,width_);
  r[width_] = '\0';
  return r;
}

Vector FitsBinColumnA::dimension()
{
  return (hastlmin_ || hastlmax_) ? Vector(tlmin_,tlmax_) : 
    Vector(DBL_MIN,DBL_MAX);
}

FitsBinColumnB::FitsBinColumnB(FitsHead* head, int i, int offset)
  : FitsBinColumn(head, i, offset)
{
  byteswap = ftlsb();

  // parse TFORM

  if (tform_) {
    istrstream str(tform_);
    if (isalpha(tform_[0]))
      str >> type_;
    else
      str >> repeat_ >> type_;
  }
}

FitsBinColumnBit::FitsBinColumnBit(FitsHead* head, int i, int off)
  : FitsBinColumnB(head, i, off)
{
  // and determine field width

  width_ = (repeat_+7)/8;
}

Vector FitsBinColumnBit::dimension()
{
  return Vector(0,repeat_);
}

template<class T> FitsBinColumnT<T>::FitsBinColumnT(FitsHead* head,
						    int i, int off)
  : FitsBinColumnB(head, i, off)
{
  // and determine field width

  width_ = repeat_ * size();
}

unsigned char FitsBinColumnT<unsigned char>::Tmin() {return 0;}
unsigned char FitsBinColumnT<unsigned char>::Tmax() {return UCHAR_MAX;}

short FitsBinColumnT<short>::Tmin() {return SHRT_MIN;}
short FitsBinColumnT<short>::Tmax() {return SHRT_MAX;}

unsigned short FitsBinColumnT<unsigned short>::Tmin() {return 0;}
unsigned short FitsBinColumnT<unsigned short>::Tmax() {return USHRT_MAX;}

#if defined( __alpha) || defined(__mips64)

int FitsBinColumnT<int>::Tmin() {return INT_MIN;}
int FitsBinColumnT<int>::Tmax() {return INT_MAX;}

unsigned int FitsBinColumnT<unsigned int>::Tmin() {return 0;}
unsigned int FitsBinColumnT<unsigned int>::Tmax() {return UINT_MAX;}

#else

long FitsBinColumnT<long>::Tmin() {return LONG_MIN;}
long FitsBinColumnT<long>::Tmax() {return LONG_MAX;}

unsigned long FitsBinColumnT<unsigned long>::Tmin() {return 0;}
unsigned long FitsBinColumnT<unsigned long>::Tmax() {return ULONG_MAX;}

#endif

float FitsBinColumnT<float>::Tmin() {return FLT_MIN;}
float FitsBinColumnT<float>::Tmax() {return FLT_MAX;}

double FitsBinColumnT<double>::Tmin() {return DBL_MIN;}
double FitsBinColumnT<double>::Tmax() {return DBL_MAX;}

unsigned char FitsBinColumnT<unsigned char>::swap(unsigned char* ptr)
{
  return *ptr;
}

short FitsBinColumnT<short>::swap(short* ptr)
{
  short r;
  swap2((char*)ptr, (char*)&r);
  return r;
}

unsigned short FitsBinColumnT<unsigned short>::swap(unsigned short* ptr)
{
  unsigned short r;
  swap2((char*)ptr, (char*)&r);
  return r;
}

BITPIX32 FitsBinColumnT<BITPIX32>::swap(BITPIX32* ptr)
{
  BITPIX32 r;
  swap4((char*)ptr, (char*)&r);
  return r;
}

unsigned BITPIX32 FitsBinColumnT<unsigned BITPIX32>::swap(unsigned BITPIX32* ptr)
{
  unsigned BITPIX32 r;
  swap4((char*)ptr, (char*)&r);
  return r;
}

float FitsBinColumnT<float>::swap(float* ptr)
{
  float r;
  swap4((char*)ptr, (char*)&r);
  return r;
}

double FitsBinColumnT<double>::swap(double* ptr)
{
  double r;
  swap8((char*)ptr, (char*)&r);
  return r;
}

double FitsBinColumnT<BITPIX32>::value(const char* ptr)
{
  union {
    char c[4];
    BITPIX32 l;
  } u;
  memcpy(u.c,ptr+offset_,4);
  return !byteswap ? u.l : swap(&u.l);
}

double FitsBinColumnT<unsigned BITPIX32>::value(const char* ptr)
{
  union {
    char c[4];
    unsigned BITPIX32 l;
  } u;
  memcpy(u.c,ptr+offset_,4);
  return !byteswap ? u.l : swap(&u.l);
}

double FitsBinColumnT<float>::value(const char* ptr)
{
  union {
    char c[4];
    float f;
  } u;
  memcpy(u.c,ptr+offset_,4);
  return !byteswap ? u.f : swap(&u.f);
}

double FitsBinColumnT<double>::value(const char* ptr)
{
  union {
    char c[8];
    double d;
  } u;
  memcpy(u.c,ptr+offset_,8);
  return !byteswap ? u.d : swap(&u.d);
}

// We need this to insure proper alignment

template<class T> double FitsBinColumnT<T>::value(const char* ptr)
{
  return !byteswap ? *((T*)(ptr+offset_)) : swap(((T*)(ptr+offset_)));
}

template<class T> double FitsBinColumnT<T>::value(const char* ptr,int i)
{
  return !byteswap ? *((T*)(ptr+offset_)+i) : swap(((T*)(ptr+offset_))+i);
}

template<class T> Vector FitsBinColumnT<T>::dimension()
{
  return (hastlmin_ || hastlmax_) ? Vector(tlmin_,tlmax_) : Vector(Tmin(),Tmax());
}

template class FitsBinColumnT<unsigned char>;
template class FitsBinColumnT<short>;
template class FitsBinColumnT<unsigned short>;
template class FitsBinColumnT<BITPIX32>;
template class FitsBinColumnT<unsigned BITPIX32>;
template class FitsBinColumnT<float>;
template class FitsBinColumnT<double>;
