#include "point.h"
#include "frame.h"
#include "util.h"

#define POINTSIZE 11

Point::Point(Frame* p, const Vector& ctr, 
	     const char* clr,  const char* f, const char* t, 
	     unsigned short prop)
  : Marker(p, ctr, clr, f, t, prop)
{
  handle = new Vector[4];
  numHandle = 4;
  gcxor = parent->markerGCXORPoint;

  updateBBox();
}

void Point::updateBBox()
{
  // bound marker

  Vector h = Vector(POINTSIZE, POINTSIZE)/2;
  Vector c = center * parent->refToCanvas;
  bbox = BBox(c-h,c+h);

  // generate handles

  handle[0] = bbox.ll;
  handle[1] = bbox.lr();
  handle[2] = bbox.ur;
  handle[3] = bbox.ul();

  // make room for handles

  bbox.expand(3);

  // calculate overall bbox

  calcAllBBox();
}

void Point::draw()
{
  Marker::draw();
  includeExclude();
  render(parent->pixmap, parent->refToWidget, gc);
}

void Point::drawXOR()
{
  render(Tk_WindowId(parent->tkwin), parent->refToWindow, gcxor);
}

void Point::drawMagnifier(const Matrix& m)
{
  XSetForeground(display, gc, color);
  includeExclude();
  render(parent->magnifierPixmap, m, gc);
}

void Point::dash(GC ggc)
{
  char dlist[] = {4,2};
  XSetLineAttributes(display, ggc, 1, LineOnOffDash, CapButt, JoinMiter);
  XSetDashes(display, ggc, 0, dlist, 2);
}

void Point::psDash()
{
  char buf[256];
  ostrstream str(buf,256);
  str << "[4 2] 0 setdash" << endl << ends;
  Tcl_AppendResult(parent->interp, buf, NULL);
}

CirclePoint::CirclePoint(Frame* p, const Vector& ctr, const char* clr,  
			 const char* f, const char* t, unsigned short prop)
  : Point(p, ctr, clr, f, t, prop)
{
  strcpy(type,"circle point");
}

void CirclePoint::render(Drawable drawable, const Matrix& m, GC ggc)
{
  Vector v = (center * m * Translate(-Vector(POINTSIZE,POINTSIZE)/2)).round();
  XDrawArc(display, drawable, ggc, (int)v[0], (int)v[1], 
	   POINTSIZE, POINTSIZE, 0, 360*64);
}

void CirclePoint::ps(int mode)
{
  Marker::ps(mode);
  psIncludeExclude();

  Vector c = center  * parent->refToCanvas;

  char buf[64];
  ostrstream str(buf,64);
  str << "newpath " << c.TkCanvasPs(parent->canvas) << POINTSIZE/2.
      << " 0 360 arc stroke" << endl << ends;
  Tcl_AppendResult(parent->interp, buf, NULL);
}

BoxPoint::BoxPoint(Frame* p, const Vector& ctr, const char* clr, const char* f,
		   const char* t, unsigned short prop) 
  : Point(p, ctr, clr, f, t, prop)
{
  strcpy(type, "box point");
}

void BoxPoint::render(Drawable drawable, const Matrix& m, GC ggc)
{
  Vector v = (center * m * Translate(-Vector(POINTSIZE,POINTSIZE)/2)).round();
  XDrawRectangle(display, drawable, ggc, (int)v[0], (int)v[1], 
		 POINTSIZE, POINTSIZE);
}

void BoxPoint::ps(int mode)
{
  Marker::ps(mode);
  psIncludeExclude();

  Vector v = center*parent->refToCanvas - Vector(POINTSIZE,-POINTSIZE)/2;

  char buf[256];
  ostrstream str(buf,256);
  str << "newpath " << v.TkCanvasPs(parent->canvas) 
      << Vector(POINTSIZE,POINTSIZE)
      << " rectstroke" << endl << ends;
  Tcl_AppendResult(parent->interp, buf, NULL);
}

DiamondPoint::DiamondPoint(Frame* p, const Vector& ctr, const char* clr, 
			   const char* f, const char* t, unsigned short prop)
  : Point(p, ctr, clr, f, t, prop) 
{
  strcpy(type, "diamond point");
}

void DiamondPoint::render(Drawable drawable, const Matrix& m, GC ggc)
{
  int l = POINTSIZE-1;
  Vector c = center * m;
  Vector v1 = (c - Vector(l,0)/2).round();
  Vector v2 = (c - Vector(0,l)/2).round();
  Vector v3 = (c + Vector(l,0)/2).round();
  Vector v4 = (c + Vector(0,l)/2).round();

  XPoint p[5];
  p[0].x = (int)v1[0];
  p[0].y = (int)v1[1];
  p[1].x = (int)v2[0];
  p[1].y = (int)v2[1];
  p[2].x = (int)v3[0];
  p[2].y = (int)v3[1];
  p[3].x = (int)v4[0];
  p[3].y = (int)v4[1];
  p[4] = p[0];

  XDrawLines(display, drawable, ggc, p, 5, CoordModeOrigin);
}

void DiamondPoint::ps(int mode)
{
  Marker::ps(mode);
  psIncludeExclude();

  Matrix m = parent->refToCanvas;

  Vector v1 = center*m - Vector(POINTSIZE, 0)/2;
  Vector v2 = center*m - Vector(0, POINTSIZE)/2;
  Vector v3 = center*m + Vector(POINTSIZE, 0)/2;
  Vector v4 = center*m + Vector(0, POINTSIZE)/2;

  char buf[512];
  ostrstream str(buf,512);
  str << "newpath " << v1.TkCanvasPs(parent->canvas) << " moveto" << endl
      << v2.TkCanvasPs(parent->canvas) << " lineto" << endl
      << v3.TkCanvasPs(parent->canvas) << " lineto" << endl
      << v4.TkCanvasPs(parent->canvas) << " lineto" << endl
      << "closepath stroke" << endl
      << ends;
  Tcl_AppendResult(parent->interp, buf, NULL);
}

CrossPoint::CrossPoint(Frame* p, const Vector& ctr, const char* clr, 
		       const char* f, const char* t, unsigned short prop) 
  : Point(p, ctr, clr, f, t, prop) 
{
  strcpy(type, "cross point");
}

void CrossPoint::render(Drawable drawable, const Matrix& m, GC ggc)
{
  Vector c = center * m;
  Vector v1 = (c - Vector(POINTSIZE,0)/2).round();
  Vector v2 = (c + Vector(POINTSIZE,0)/2).round();
  Vector v3 = (c - Vector(0,POINTSIZE)/2).round();
  Vector v4 = (c + Vector(0,POINTSIZE)/2).round();

  XDrawLine(display, drawable, ggc, (int)v1[0], (int)v1[1], 
	    (int)v2[0], (int)v2[1]);
  XDrawLine(display, drawable, ggc, (int)v3[0], (int)v3[1], 
	    (int)v4[0], (int)v4[1]);
}

void CrossPoint::ps(int mode)
{
  Marker::ps(mode);
  psIncludeExclude();

  Matrix m = parent->refToCanvas;

  Vector v1 = center*m - Vector(POINTSIZE,0)/2;
  Vector v2 = center*m + Vector(POINTSIZE,0)/2;
  Vector v3 = center*m - Vector(0,POINTSIZE)/2;
  Vector v4 = center*m + Vector(0,POINTSIZE)/2;

  char buf[256];
  ostrstream str(buf,256);
  str << "newpath " << v1.TkCanvasPs(parent->canvas) << " moveto" << endl
      << v2.TkCanvasPs(parent->canvas) << " lineto stroke" << endl
      << "newpath " << v3.TkCanvasPs(parent->canvas) << " moveto" << endl
      << v4.TkCanvasPs(parent->canvas) << " lineto stroke" << endl
      << ends;
  Tcl_AppendResult(parent->interp, buf, NULL);
}

ExPoint::ExPoint(Frame* p, const Vector& ctr, const char* clr,  const char* f,
		 const char* t, unsigned short prop) 
  : Point(p, ctr, clr, f, t, prop) 
{
  strcpy(type, "x point");
}

void ExPoint::render(Drawable drawable, const Matrix& m, GC ggc)
{
  Vector c = center * m;
  Vector v1 = (c + Vector(-POINTSIZE,-POINTSIZE)/2).round();
  Vector v2 = (c + Vector( POINTSIZE,-POINTSIZE)/2).round();
  Vector v3 = (c + Vector( POINTSIZE, POINTSIZE)/2).round();
  Vector v4 = (c + Vector(-POINTSIZE, POINTSIZE)/2).round();

  XDrawLine(display, drawable, ggc, (int)v1[0], (int)v1[1], 
	    (int)v3[0], (int)v3[1]);
  XDrawLine(display, drawable, ggc, (int)v2[0], (int)v2[1], 
	    (int)v4[0], (int)v4[1]);
}

void ExPoint::ps(int mode)
{
  Marker::ps(mode);
  psIncludeExclude();

  Matrix m = parent->refToCanvas;

  Vector v1 = center*m + Vector(-POINTSIZE, -POINTSIZE)/2;
  Vector v2 = center*m + Vector(POINTSIZE, -POINTSIZE)/2;
  Vector v3 = center*m + Vector(POINTSIZE, POINTSIZE)/2;
  Vector v4 = center*m + Vector(-POINTSIZE, POINTSIZE)/2;

  char buf[256];
  ostrstream str(buf,256);
  str << "newpath " << v1.TkCanvasPs(parent->canvas) << " moveto" << endl
      << v3.TkCanvasPs(parent->canvas) << " lineto stroke" << endl
      << "newpath " << v2.TkCanvasPs(parent->canvas) << " moveto" << endl
      << v4.TkCanvasPs(parent->canvas) << " lineto stroke" << endl
      << ends;
  Tcl_AppendResult(parent->interp, buf, NULL);
}

ArrowPoint::ArrowPoint(Frame* p, const Vector& ctr, const char* clr, 
		       const char* f, const char* t, unsigned short prop) 
  : Point(p, ctr, clr, f, t, prop) 
{
  strcpy(type, "arrow point");
}

void ArrowPoint::render(Drawable drawable, const Matrix& m, GC ggc)
{
  Vector c = center * m;
  Vector v1 = (c + Vector(-POINTSIZE, 0)/2).round();
  Vector v2 = (c + Vector(0, -POINTSIZE)/2).round();
  Vector v3 = (c - Vector(POINTSIZE, POINTSIZE)/2).round();
  Vector v4 = c.round();

  XDrawLine(display, drawable, ggc, (int)v1[0], (int)v1[1], 
	    (int)v4[0], (int)v4[1]);
  XDrawLine(display, drawable, ggc, (int)v2[0], (int)v2[1], 
	    (int)v4[0], (int)v4[1]);
  XDrawLine(display, drawable, ggc, (int)v3[0], (int)v3[1], 
	    (int)v4[0], (int)v4[1]);
}

void ArrowPoint::ps(int mode)
{
  Marker::ps(mode);
  psIncludeExclude();

  Matrix m = parent->refToCanvas;

  Vector v1 = center*m + Vector(-POINTSIZE, 0)/2;
  Vector v2 = center*m + Vector(0, -POINTSIZE)/2;
  Vector v3 = center*m - Vector(POINTSIZE, POINTSIZE)/2;
  Vector v4 = center*m;

  char buf[512];
  ostrstream str(buf,512);
  str << "newpath " << v1.TkCanvasPs(parent->canvas) << " moveto" << endl
      << v4.TkCanvasPs(parent->canvas) << " lineto stroke" << endl
      << "newpath " << v2.TkCanvasPs(parent->canvas) << " moveto" << endl
      << v4.TkCanvasPs(parent->canvas) << " lineto stroke" << endl
      << "newpath " << v3.TkCanvasPs(parent->canvas) << " moveto" << endl
      << v4.TkCanvasPs(parent->canvas) << " lineto stroke" << endl
      << ends;
  Tcl_AppendResult(parent->interp, buf, NULL);
}

BoxCirclePoint::BoxCirclePoint(Frame* p, const Vector& ctr, const char* clr, 
			       const char* f, const char* t, 
			       unsigned short prop)
  : Point(p, ctr, clr, f, t, prop) 
{
  strcpy(type, "boxcircle point");
}

void BoxCirclePoint::render(Drawable drawable, const Matrix& m, GC ggc)
{
  Vector c = center * m;
  Vector v1 = (c * Translate(-Vector(POINTSIZE-2,POINTSIZE-2)/2)).round();
  Vector v2 = (c * Translate(-Vector(POINTSIZE,POINTSIZE)/2)).round();

  XDrawArc(display, drawable, ggc, (int)v1[0], (int)v1[1], 
	   POINTSIZE-2, POINTSIZE-2, 0, 360*64);
  XDrawRectangle(display, drawable, ggc, (int)v2[0], (int)v2[1], 
		 POINTSIZE, POINTSIZE);
}

void BoxCirclePoint::ps(int mode)
{
  Marker::ps(mode);
  psIncludeExclude();

  Vector v1 = center*parent->refToCanvas;
  Vector v2 = center*parent->refToCanvas - Vector(POINTSIZE,-POINTSIZE)/2;

  char buf[256];
  ostrstream str(buf,256);
  str << "newpath " << v1.TkCanvasPs(parent->canvas) << (POINTSIZE-2)/2.
      << " 0 360 arc stroke" << endl
      << "newpath " << v2.TkCanvasPs(parent->canvas) 
      << Vector(POINTSIZE,POINTSIZE) 
      << " rectstroke" << endl
      << ends;
  Tcl_AppendResult(parent->interp, buf, NULL);
}

// list

void Point::listSAOtng(ostream& str, int sys, int format)
{
  if (text[0])
    str << '#' << text << endl;

  if (properties&INCLUDE)
    str << '+';
  else
    str << '-';

  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector v = parent->map(center,Frame::REF,Frame::IMAGE);
      str << "point(" << setprecision(8) << v[0] << ',' << v[1] << ") ";
    }
    break;
  case Frame::PHYSICAL:
    {
      Vector v = parent->map(center,Frame::REF,Frame::PHYSICAL);
      str << "point(" << setprecision(8) << v[0] << ',' << v[1] << ") ";
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::LINEAR:
  case Frame::ICRS:
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	str << "point(" << setprecision(8) << v[0] << ',' << v[1] << ") ";
      }
      break;
    case Frame::SEXAGESIMAL:
      {
	char buf[64];
	parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			    (Frame::CoordFormat)format, buf, 64);
	char ra[16];
	char dec[16];
	istrstream wcs(buf,64);
	wcs >> ra >> dec;
	str << "point(" << ra << ',' << dec << ") ";
      }
      break;
    }
    break;
  }

  str << "# " << colorName;
}

void Point::listPros(ostream& str, int sys, int format)
{
  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector v = parent->map(center,Frame::REF,Frame::IMAGE);
      str << "point " << setprecision(8) << v;
    }
    break;

  case Frame::PHYSICAL:
    {
      str << "physical; ";
      Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
      str << "point " << setprecision(8) << v;
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::LINEAR:
  case Frame::ICRS:
    parent->listProsCoordSystem(str, (Frame::CoordSystem)sys);
    str << "; ";
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	str << "point " << setprecision(8) << v[0] << "d " << v[1] << "d ";
      }
      break;
    case Frame::SEXAGESIMAL:
      {
	char buf[64];
	parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			    (Frame::CoordFormat)format, buf, 64);
	char ra[16];
	char dec[16];
	istrstream wcs(buf,64);
	wcs >> ra >> dec;
	if (dec[0]=='+')
	  str << "point " << ra << ' ' << dec+1;
	else
	  str << "point " << ra << ' ' << dec;
      }
      break;
    }
    break;
  }
}

void Point::listSAOimage(ostream& str)
{
  // all coords are in image coords

  Vector v = parent->map(center,Frame::REF,Frame::IMAGE);
  str << "point(" << setprecision(8) << v[0] << ',' << v[1] << ')';
}
